/*----------------------------------------------------------------------*/
/*      msortv.c        test sort                                       */
/*                                                                      */
/*      Jeff Reid       13AUG15  14:00                                  */
/*----------------------------------------------------------------------*/
/*      get rid of warnings for fopen */
#define _CRT_SECURE_NO_WARNINGS 1

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define NUMREC (1024*4096)          /* number of records */

typedef unsigned __int64    UI64;
typedef unsigned __int64 *  PUI64;

#define CMPFUNC 0                   /* set to 1 to use compare function */

#if CMPFUNC
typedef int (*PFCMP)(UI64, UI64);
static int  ElemCmp(UI64, UI64);
PFCMP pfCmp = ElemCmp;
static UI64 qwCmp;
#endif

static clock_t dwTimeStart;             /* clock values */
static clock_t dwTimeStop;

static FILE * fpDst;                    /* dest file */

static void  MakeGroups(PUI64);
static PUI64 MergeGroups(PUI64, PUI64);

int main( )
{
PUI64 pData = NULL;
PUI64 pTemp = NULL;
PUI64 pSort = NULL;
UI64 i, r;

/*  allocate space of data and temp arrays */

    pData = malloc(NUMREC*sizeof(UI64));
    if(pData == NULL)
        goto exit0;
    pTemp = malloc(NUMREC*sizeof(UI64));
    if(pTemp == NULL)
        goto exit0;

/*      generate data */

    for(i = 0; i < NUMREC; i++){
        r  = (((UI64)((rand()>>4) & 0xff))<< 0);
        r += (((UI64)((rand()>>4) & 0xff))<< 8);
        r += (((UI64)((rand()>>4) & 0xff))<<16);
        r += (((UI64)((rand()>>4) & 0xff))<<24);
        r += (((UI64)((rand()>>4) & 0xff))<<32);
        r += (((UI64)((rand()>>4) & 0xff))<<40);
        r += (((UI64)((rand()>>4) & 0xff))<<48);
        r += (((UI64)((rand()>>4) & 0xff))<<56);
        pData[i] = r;
    }

/*      sort data */

    dwTimeStart = clock();

    MakeGroups(pData);

    pSort = MergeGroups(pData, pTemp);

    dwTimeStop = clock();
    printf("Number of ticks    %d\n", dwTimeStop-dwTimeStart);

#if CMPFUNC
    printf("Number of compares %d\n", qwCmp);
#endif

/*      output data */

    fpDst = fopen("dstm.bin", "wb");
    if(fpDst == NULL){
        goto exit0;}
    fwrite(pSort, 1, NUMREC*sizeof(UI64), fpDst);
    fclose(fpDst);

exit0:
    if(pData)
        free(pData);
    if(pTemp)
        free(pTemp);
    return(0);
}

#if CMPFUNC
/*----------------------------------------------------------------------*/
/*      ElemCmp     element compare                                     */
/*----------------------------------------------------------------------*/
static int ElemCmp(UI64 elem1, UI64 elem2)
{
    qwCmp += 1;                         /* bump # of compares */
    return elem1 < elem2;               /* return compare value */
}
#endif


/*----------------------------------------------------------------------*/
/*      MakeGroups                                                      */
/*----------------------------------------------------------------------*/
static void MakeGroups(PUI64 pv)
{
PUI64   pe;
UI64    qw0, qw1;

    pe = pv + (NUMREC&0xfffffffeul);
    while(pv < pe){
        qw0 = pv[0];
        qw1 = pv[1];
#if CMPFUNC
        if(pfCmp(qw1, qw0)){
#else
        if(qw1 < qw0){
#endif
            pv[0] = qw1;
            pv[1] = qw0;}
        pv += 2;}
}


/*----------------------------------------------------------------------*/
/*      MergeGroups                                                     */
/*----------------------------------------------------------------------*/
static PUI64 MergeGroups(PUI64 pv0, PUI64 pv1)
{
PUI64   pGet;                           /* current pass base ptrs */
PUI64   pPut;
PUI64   pGet0;                          /* working ptrs */
PUI64   pGet1;
PUI64   pEnd;
PUI64   pEnd0;
PUI64   pEnd1;
PUI64   pPutx;
UI64    qwRec0;                         /* working records */
UI64    qwRec1;
UI64    qwGrpCnt;                       /* group size */

    pGet = pv0;                         /* set up for first pass */
    pPut = pv1;
    qwGrpCnt = 2;

/*      start of a pass */

    while(qwGrpCnt < NUMREC){           /* while not done */
        pGet1 = pGet;                   /* update pass ptrs */
        pPutx = pPut;
        pEnd  = pGet+NUMREC;

/*      merge two groups */

        while(1){
            pGet0 = pGet1;              /* update group ptrs */
            pEnd0 = pGet0 + qwGrpCnt;
            if(pEnd0 >= pEnd)
                pEnd0 = pEnd;
            pGet1 = pEnd0;
            pEnd1 = pGet1 + qwGrpCnt;
            if(pEnd1 >= pEnd)
                pEnd1 = pEnd;
            if(pGet0 == pEnd0)          /* if no data, end pass */
                break;
            if(pGet1 == pEnd1){         /* if only group0, copy it, end pass */
                while(pGet0 != pEnd0)
                    *pPutx++ = *pGet0++;
                break;}

/*      repeatedly copy the lower of two records */
/*      until both groups copied */

            qwRec0 = *pGet0++;          /* get 1st 2 records */
            qwRec1 = *pGet1++;
            while(1){
#if CMPFUNC
                if(!pfCmp(qwRec1, qwRec0)){ /* if 0 <= 1, put rec0 */
#else
                if(!(qwRec1 < qwRec0)){ /* if 0 <= 1, put rec0 */
#endif
                    *pPutx++ = qwRec0;
                    if(pGet0 < pEnd0){  /*   if not end 0, get next rec0 */
                        qwRec0 = *pGet0++;
                        continue;}
                    else{               /*   else copy rest of grp1 */
                        *pPutx++ = qwRec1;
                        while(pGet1 != pEnd1)
                            *pPutx++ = *pGet1++;
                        break;}}        /*   end two group merge */
                else{                   /* else 1 < 0, put rec1 */
                    *pPutx++ = qwRec1;
                    if(pGet1 < pEnd1){  /*   if not end 1, get next rec1 */
                        qwRec1 = *pGet1++;
                        continue;}
                    else{               /*   else copy rest of grp0 */
                        *pPutx++ = qwRec0;
                        while(pGet0 != pEnd0)
                            *pPutx++ = *pGet0++;
                        break;}}}}      /*   end two group merge */
        pEnd = pPut;                    /* swap ptrs */
        pPut = pGet;
        pGet = pEnd;
        qwGrpCnt <<= 1;}                /* bump group size */

    return(pGet);                       /* return ptr to sorted data */
}
